%{
  open Cnf;;

  let md5s s = Int64.to_int ((Obj.magic (Digest.to_hex (Digest.string s))) : int64);;

  type lit = int * term list
  type form = Atom of (int * term list)
            | Neg of form
            | Conj of form * form
            | Disj of form * form
            | Forall of int * form
            | Exists of int * form
            | Eqiv of form * form;;

  let cnst_no, no_cnst = Hashtbl.create 100, Hashtbl.create 100;;

  let rec find_free_const n = if Hashtbl.mem no_cnst n then find_free_const (n + 1) else n;;
  let rec find_free_pred n = 
    if Hashtbl.mem no_cnst n || Hashtbl.mem no_cnst (-n) then
      find_free_pred (if n + 1 < 0 then 1 else n + 1) else n;;

  let const_number name =
    try Hashtbl.find cnst_no name with Not_found ->
      let cno = find_free_pred (abs (md5s name)) in
      Hashtbl.add cnst_no name cno; Hashtbl.add no_cnst cno name; cno;;

  let pred_number name =
    try Hashtbl.find cnst_no name with Not_found ->
      let pno = find_free_pred (abs (md5s name)) in
      Hashtbl.add cnst_no name pno; Hashtbl.add no_cnst pno name; pno;;

  let const name args = A (const_number name, args);;
  let pred name args = Atom (pred_number name, args);;

  let var_no, no_var, var_num = Hashtbl.create 100, Hashtbl.create 100, ref 0;;

  let var (name : string) =
    try Hashtbl.find var_no name with Not_found ->
      incr var_num; Hashtbl.add var_no name !var_num; Hashtbl.add no_var !var_num name; !var_num;;

  let eqn = pred_number "=";;

  let list_forall vs t = List.fold_right (fun v sf -> Forall (var v, sf)) vs t;;
  let list_exists vs t = List.fold_right (fun v sf -> Exists (var v, sf)) vs t;;
  let list_conj = function [] -> invalid_arg "list_conj"
    | h :: t -> List.fold_left (fun sf e -> Conj (sf, e)) h t;;
  let list_disj = function [] -> invalid_arg "list_disj"
    | h :: t -> List.fold_left (fun sf e -> Disj (sf, e)) h t;;
  let is_uppercase s = s <> "" && s.[0] <> Char.lowercase s.[0];;

%}

%token <string> Word
%token <string list> All
%token <string list> Ex
%token Eof Openp Closep Dot Comma Eq Neq Tilde Eqvt Neqvt Impl Revimpl And Or
%right Impl
%nonassoc Eqvt
%nonassoc Eq Neq
%right Or
%left And
%nonassoc Tilde
%nonassoc All Exists

%start fof_top fof_data
%type <(string * string * form)> fof_top
%type <(form * form * form list * form list)> fof_data
%%

fof_top :
  Word Openp Word Comma Word Comma formula Closep Dot { ($3, $5, $7) }
| Eof { raise End_of_file };

formula :
| qformula     { $1 }
| formula Impl formula { Disj (Neg $1, $3) }
| formula Revimpl formula { Disj (Neg $3, $1) }
| formula Eqvt formula  { Eqiv ($1, $3) }
| formula And formula { Conj ($1, $3) }
| formula Or formula { Disj ($1, $3) }
| formula Neqvt formula { Neg (Eqiv ($1, $3)) }

qformula :
  Word { pred $1 [] }
| Word Openp ts { pred $1 $3 }
| fterm Eq fterm { Atom (eqn, [$1; $3]) }
| fterm Neq fterm { Neg (Atom (eqn, [$1; $3])) }
| Tilde qformula { Neg $2 }
| All qformula { list_forall $1 $2 }
| Ex qformula { list_exists $1 $2 }
| Openp formula Closep { $2 }

fterm :
  Word { if is_uppercase $1 then V (var $1) else const $1 [] }
| Word Openp ts { const $1 $3 }

ts :
  fterm Comma ts { $1 :: $3 }
| fterm Closep   { [$1] };

flist :
  formula Comma flist  { $1 :: $3 }
| formula Closep { [$1] }
| Closep { [] };

fof_data :
  formula Comma formula Comma Openp flist Comma Openp flist Dot { ($1, $3, $6, $9) }
| Eof { raise End_of_file };
